proc balloon_create {w {genCmd ""}} {
	global balloonInfo
	bind $w <Enter> "balloon_poll $w"
	bind $w <Leave> "balloon_deactivate $w"
	bind $w <Destroy> "balloon_deactivate $w"
	bind $w <Motion> "balloon_deactivate $w"
	if {![info exists balloonInfo(label)]} {
		set l [toplevel .ballonhelp  -highlightbackground black \
			-highlightthickness 1]
		pack [label $l.l -background #FFFFBF]
		wm overrideredirect $l 1
		set balloonInfo(label) $l
		catch {$l configure -font -Adobe-Helvetica-Medium-R-Normal--*-120-*-*-*-*-*-*}
		wm withdraw $l
		set balloonInfo(xLast) -1
		set balloonInfo(yLast) -1
		set balloonInfo(noshow) 0
	}
	set balloonInfo(gen:$w) $genCmd
	set balloonInfo(text:$w) ""
}

proc balloon_poll {w} {
	global balloonInfo
	set x [winfo pointerx $w]
	set y [winfo pointery $w]
	if {($x==$balloonInfo(xLast))&&($y==$balloonInfo(yLast))} {
		if {$balloonInfo(noshow)} {
			set balloonInfo(event) [after 500 "balloon_poll $w"]
			return
		}
		set l $balloonInfo(label)
		eval $balloonInfo(gen:$w)
		if {$balloonInfo(text:$w) == ""} {
			balloon_deactivate $w
			set balloonInfo(noshow) 1
			balloon_poll $w
		} else {
			$l.l configure -text $balloonInfo(text:$w)
			wm geometry $l +[expr $x+7]+[expr $y+10]
			wm deiconify $l
			raise $l
			# keep displaying tag until cursor leaves target widget
			bind $w <Motion> ""
			set balloonInfo(active) $w
		}
	} else {
		set balloonInfo(noshow) 0
		set balloonInfo(xLast) $x
		set balloonInfo(yLast) $y
		set balloonInfo(event) [after 500 "balloon_poll $w"]
		set balloonInfo(active) $w
	}
}

proc balloon_deactivate {w} {
	global balloonInfo
	catch {after cancel $balloonInfo(event)}
	catch {unset balloonInfo(active)}
	wm withdraw $balloonInfo(label)
}

proc balloon_setmessage {w msg} {
	global balloonInfo
	set balloonInfo(text:$w) $msg
}
set BUTTONS_TYPE ""
proc buttons_create {w args} {
	global BUTTONS_TYPE
	set i 1

	frame $w
	foreach l $args {
		set b b$i
		set name [lindex [split $l :] 0]
		set opt  [lindex [split $l :] 1]
		if [string match $opt ""] {set opt l}
		set flag [options_translate $opt]
		if {$BUTTONS_TYPE == ""} {
			button $w.$b -text $name
		} else {
			$BUTTONS_TYPE $w.$b -text $name
		}
		eval pack $w.$b $flag
		incr i
	}
	buttons_default $w 1
}

proc buttons_add {w l} {
	set name [lindex [split $l :] 0]
	set opt  [lindex [split $l :] 1]
	set i 1
	if [string match $opt ""] {set opt l}
	while {[winfo exists $w.b$i]} {
		incr i
	}
	set flag [options_translate $opt]
	button $w.b$i -text $name
	eval pack $w.b$i $flag
	return $i
}

proc buttons_command {w num cmd} {
	set b $w.b$num
	set res [$b cget -command]
	$b config -command $cmd
	return $res
}

proc buttons_default {w num} {
	focus $w.b$num
}

proc dialog_entry {msg {title "Entry Box"} {initText ""}} {
	global dialogEntrySet
	set dialogEntrySet ""
	set w [window_create]
	window_title $w $title
	label_create $w.l $msg
	entry $w.e -width 40
	$w.e insert end $initText
	bind $w.e <Return> "set dialogEntrySet 1"
	buttons_create $w.b OK:r Cancel:r
	buttons_command $w.b 1 "set dialogEntrySet 1"
	buttons_command $w.b 2 "set dialogEntrySet 0"
	mpack $w.l:tx $w.e:tze $w.b:bx
	window_grab $w
	focus $w.e
	vwait dialogEntrySet
	window_ungrab $w
	set e [$w.e get]
	destroy $w
	if {$dialogEntrySet} {return $e}
}

proc dialog_text {msg {title "Text Entry Box"}} {
	global dialogEntrySet
	set dialogEntrySet ""
	set w [window_create]
	window_title $w $title
	label_create $w.l $msg
	text $w.e -width 40 -height 5
	buttons_create $w.b OK:r Cancel:r
	buttons_command $w.b 1 "set dialogEntrySet 1"
	buttons_command $w.b 2 "set dialogEntrySet 0"
	mpack $w.l:tx $w.e:tze $w.b:bx
	window_grab $w
	focus $w.e
	vwait dialogEntrySet
	window_ungrab $w
	set e [$w.e get 0.0 end]
	destroy $w
	if {$dialogEntrySet} {return $e}
}
proc entry_create {w label {button ""}} {
	frame $w
	label $w.l -text $label
        entry $w.e -background white
        pack $w.l -side left -fill x
	if {! [string match $button ""]} {
		button $w.b -text $button
		pack $w.b -side right
	}
        pack $w.e -side left -expand 1 -fill x
	return $w
}

proc entry_cmd {w command} {
        $w.b config -command $command
}

proc entry_get {w} {
	return [$w.e get]
}

proc entry_set {w newval} {
	$w.e delete 0 end
	$w.e insert 0 $newval
}

global GetFileName GetFileDone

proc getfile_open {{msg "Select a file:"} {f ""}} {
	global tcl_version GetFileName
	if {$tcl_version > 7.5} {
		set GetFileName [tk_getOpenFile -title $msg]
		return $GetFileName
	} else {
		return [getfile_generic $msg $f]
	}
}

proc getfile_save {{msg "Select a file:"} {f ""}} {
	global tcl_version GetFileName
	if {$tcl_version > 7.5} {
		set GetFileName [tk_getSaveFile -title $msg]
		return $GetFileName
	} else {
		return [getfile_generic $msg $f]
	}
}

proc getfile_generic {{msg "Select a file:"} {f ""}} {
	global tix_version GetFileName
	if {[info exists tix_version]} {
		set dialog [tix filedialog tixExFileSelectDialog]
		$dialog config -command {set GetFileName}
		$dialog popup
		tkwait variable GetFileName
		return $GetFileName
	} else {
		return [getfile_setup $msg $f]
	}
}

proc getfile_setup {{msg "Select a file:"} {f ""}} {
	global GetFileName GetFileDone
	if {$f == ""} {set f [pwd]/}
	set olddir [pwd]
	set w [toplevel .fileselect -bd 4 -class Getfile]
	set GetFileName "$f"

	getfile_resources
	getfile_window $w "$msg"
	getfile_setfiles $w.files

	bind $w.files.file.list "<1>" \
		"getfile_sclick %W %y ; focus $w.sel.e"
	bind $w.files.dir.list "<1>" \
		"getfile_sclick %W %y ; focus $w.sel.e"
	bind $w.files.dir.list "<Double-1>" \
		"getfile_dclick $w.files %y ; focus $w.sel.e"
	bind $w.sel.e "<Return>" "getfile_setfiles $w.files"

	tkwait variable GetFileDone
	destroy $w
	cd "$olddir"
	return $GetFileName
}

proc getfile_window {w msg} {
	global GetFileName GetFileDone
	# Create everything
	message $w.msg -aspect 1000 -text "$msg"

	frame $w.sel
	label $w.sel.l -padx 0
	entry $w.sel.e -textvariable GetFileName

	frame $w.files
	frame $w.files.dir
	frame $w.files.file
	getfile_xyscroll $w.files.dir
	getfile_xyscroll $w.files.file

	frame $w.buttons
	button $w.buttons.ok -command "set GetFileDone 1"
	button $w.buttons.cancel -command "set GetFileName {};set GetFileDone 1"

	# Pack everything

	pack $w.sel.l -side left
	pack $w.sel.e -side right -fill x -expand 1
	pack $w.msg -side top -fill x
	pack $w.files.dir $w.files.file -side left -fill both
	pack $w.buttons.ok $w.buttons.cancel -side right
	pack $w.msg $w.sel $w.files $w.buttons -side top -fill both
}

proc getfile_resources {} {
	option add "*Getfile*ok.text"		OK		startup
	option add "*Getfile*cancel.text"	Cancel		startup
	option add "*Getfile*l.text"		Selection:	startup
}

proc getfile_xyscroll {w} {
	listbox $w.list -yscrollcommand "$w.scroll set" -border 1 -relief raised
	scrollbar $w.scroll -command "$w.list yview"
	pack $w.list -side left -fill both -expand 1
	pack $w.scroll -side left -fill y
}

proc getfile_setfiles {l} {
	global GetFileName GetFileDone

	if [file isdir $GetFileName] {
		catch {cd $GetFileName}
	} {
		catch {cd [file dirname $GetFileName]}
	}
	set dwidget $l.dir.list
	set fwidget $l.file.list

	$dwidget delete 0 end
	$fwidget delete 0 end
	$dwidget insert end . ..
	foreach i [lsort [glob *]] {
		if {[file isfile $i]} {set w $fwidget} {set w $dwidget}
		$w insert end $i
	}
}

proc getfile_sclick {w p} {
	global GetFileName GetFileDone
	set GetFileName [pwd]/[$w get [$w nearest $p]]
}

proc getfile_dclick {w p} {
	global GetFileName GetFileDone
	cd [$w.dir.list get [$w.dir.list nearest $p]]
	set GetFileName [pwd]/
	getfile_setfiles $w
}
###################################################################
# hlist.tcl
#
#	Simulates a hierarchical list widget. EG:
#		pack [hlist_create .foo]
#		hlist_insert .foo a "First Entry"
#		hlist_insert .foo b "Second Entry"
#		hlist_insert .foo a/a "Subwidget of First Entry"
###################################################################

global HlistData
bind Hlist       <ButtonPress>	{hlist_selectloc %W %x %y}
bind HlistButton <ButtonPress>	{hlist_selectloc_btn [winfo parent %W] %X %Y}
bind HlistButton <Enter>	{%W config -relief sunken}
bind HlistButton <Leave>	{%W config -relief flat}
if {[string match /* [file dirname [info script]]]} {
	set HlistData(libDir) [file dirname [info script]]
} else {
	set HlistData(libDir) [pwd]/[file dirname [info script]]
}

proc hlist_selectloc {w x y} {
	set t [$w index @$x,$y]
	$w tag remove sel 0.0 end
	$w tag add sel "$t linestart+1c" "$t lineend"
}
proc hlist_selectloc_btn {w x y} {
   hlist_selectloc $w [expr $x - [winfo rootx $w]] [expr $y - [winfo rooty $w]]
   hlist_choose [winfo parent $w] [lindex [hlist_selection [winfo parent $w]] 0]
}


proc hlist_create {w} {
	global HlistData
	frame $w
	text $w.t -relief raised -yscrollcommand "$w.s set" -setgrid true \
		-font -Adobe-Helvetica-Bold-R-Normal--*-120-*-*-*-*-*-* -bd 2 \
		-cursor "" -width 45 -height 25
        scrollbar $w.s -command "$w.t yview"
        pack $w.s -side right -fill y
        pack $w.t -expand yes -fill both
	bindtags $w.t "$w.t Hlist [winfo toplevel $w.t] all"
	$w.t tag configure sel -background yellow
	set HlistData(l$w) 0
	return $w
}

proc hlist_subwidget {w name {id ""}} {
	global HlistData
	switch $name {
		text		{return $w.t}
		scrollbar	{return $w.s}
		label		{return $HlistData(l$w:$name)}
	}
}

proc hlist_insert {w name text {clientdata ""}} {
	global HlistData

	if {$clientdata != ""} {set HlistData(c$w:$name) $clientdata}
	if {[$w.t tag ranges $name] != ""} return
	set HlistData(n$w:$name) $text
	set HlistData(d$w:$name) 1
	set max 0.0
	set parent [file dirname $name]
	foreach t "n$w:$parent [array names HlistData n$w:$parent/*]" {
		regexp {:(.*)} $t dummy t
		if {[string compare $t $name] != -1} continue
		set i [lindex [$w.t tag nextrange $t 0.0] 1]
		if {$i != ""} {if {$i > $max} {set max $i}}
	}
	if {$max == "0.0"} {set max end}
	set level [regsub -all / $name / dummy]
	set margin "[expr $level * 5]m"
	$w.t tag configure $name -lmargin1 $margin -lmargin2 $margin
	$w.t insert $max " $text\n" $name
	$w.t see $max
	set l $w.t.l_$HlistData(l$w)
	incr HlistData(l$w)
	set HlistData(l$w:$name) $l
	label $l
	hlist_seticon $w $name none
	if {$parent != "." && $parent != "/"} {hlist_seticon $w $parent minus}
	if {[array names HlistData n$w:$name/*] != ""} {
		hlist_seticon $w $name plus
	}
	bindtags $l "HlistButton [bindtags $l]"
	$w.t tag bind $name <Double-1> "hlist_choose $w $name"
	$w.t tag bind $name <Double-3> "hlist_expand $w $name"
	$w.t window create $name.first+1c -window $l
}

proc hlist_delete {w name} {
	global HlistData
	catch {eval $w.t delete [$w.t tag nextrange $name 0.0]}
	catch {$w.t tag delete $name}
	catch {destroy $HlistData(l$w:$name)}
	catch {unset HlistData(n$w:$name)}
	catch {unset HlistData(c$w:$name)}
	catch {unset HlistData(l$w:$name)}
	catch {unset HlistData(d$w:$name)}
}

proc hlist_names {w {pat *}} {
	global HlistData
	set result ""
	foreach t [$w.t tag names] {
		if {[info exists HlistData(n$w:$t)] && [string match $pat $t]} {
			lappend result $t
		}
	}
	return $result
}

proc hlist_clear {w {pat *}} {
	foreach t [hlist_names $w $pat] {hlist_delete $w $t}
}

proc hlist_choose {w name} {
	if {[hlist_geticon $w $name] == "minus"} {
		hlist_collapse $w $name
	} else {
		hlist_expand $w $name 
	}
}

proc hlist_collapse {w name} {
	global HlistData
	set hasChild 0
	foreach t [$w.t tag names] {
		if {[string match $name/* $t]} {
			set start [lindex [$w.t tag nextrange $t 0.0] 0]
			$w.t delete "$start linestart" "$start lineend+1c"
			$w.t tag delete $t
			if {[file dirname $t] == $name} {set HlistData(d$w:$t) 0}
			set hasChild 1
		}
	}
	if {$hasChild} {hlist_seticon $w $name plus}
}

proc hlist_expand {w name} {
	global HlistData
	foreach t [lsort [array names HlistData n$w:$name/*]] {
		regexp {:(.*)} $t dummy n
		if {[file dirname $n] == $name} {set HlistData(d$w:$n) 1}
		if {$HlistData(d$w:$n)} {
			hlist_insert $w $n $HlistData($t)
		}
	}
	set HlistData(d$w:$name) 1
	if {[info exists HlistData(a$w:$name)]} {
		eval $HlistData(a$w:$name)
	}
}

proc hlist_seticon {w name s} {
	global HlistData
	set label $HlistData(l$w:$name)
	set libDir $HlistData(libDir)
	switch $s {
	 plus - close {$label configure -bitmap @$libDir/reg_plus.xbm}
	 minus - open {$label configure -bitmap @$libDir/reg_minus.xbm}
	 empty - none {$label configure -bitmap @$libDir/reg.xbm}
	}
}

proc hlist_geticon {w name} {
	global HlistData
	set label $HlistData(l$w:$name)
	switch -glob [$label cget -bitmap] {
		{*/reg.xbm} {return empty}
		{*/reg_plus.xbm} {return plus}
		{*/reg_minus.xbm} {return minus}
	}
}

proc hlist_action {w name {cmd ""}} {
	global HlistData
	if {$cmd != ""} {
		set HlistData(a$w:$name) $cmd
	} else {
		catch {unset HlistData(a$w:$name)}
	}
}

proc hlist_text {w t} {
	global HlistData
	return $HlistData(n$w:$t)
}

proc hlist_selection {w {t ""}} {
	global HlistData
	if {$t != ""} {
		set r ""
		catch {set r $HlistData(c$w:$t)}
		return $r
	}
	set s [$w.t tag nextrange sel 0.0]
	if {$s != ""} {
		foreach t [$w.t tag names [lindex $s 0]] {
			if {[info exists HlistData(n$w:$t)]} {
				if {[info exists HlistData(c$w:$t)]} {
					return [list $t $HlistData(c$w:$t)]
				} else {
					return $t
				}
			}
		}
	}
}
global MenuData

proc menu_create {w args} {
	global BUTTONS_TYPE
	set BUTTONS_TYPE menu_createbutton
	eval buttons_create $w $args
	set BUTTONS_TYPE ""
	$w config -relief raised -borderwidth 1
}

proc menu_createbutton {w args} {
	eval menubutton $w $args -menu $w.menu
	menu $w.menu
}

proc menu_convertname {w number} {
	regsub {\..*$} $number "" b
	regsub {^[0-9]*} $number "" rest
	regsub -all {([0-9]+)} $rest {m\1} result
	return $w.b$b.menu$result
}

proc menu_convertname_new {w number} {
	set b [shpat_%% $number {\..*}]
	set rest [shpat_## $number {[0-9]+}]
	regsub -all {([0-9]+)} $rest {m\1} result
	return "$w.b$b.menu$result"
}

proc menu_index {w n base} {
	set m [menu_convertname $w $n]
	return [$m index $base]
}

proc menu_additems {w args} {
	set prefix [menu_getassoc $w prefix]
	
	foreach v $args {
		set m 1
		set orig $v

		while {[string match */* $v]} {
			regsub {/.*$} $v "" base
			if [catch {set madeMenu($base)}] {
				set madeMenu($base) ""
			}
			if [string match "" $madeMenu($base)] {
				menu_addcascade $w $m $base
				set madeMenu($base) $m.[menu_index $w $m $base]
			}
			set m $madeMenu($base)
			regsub {^[^/]*/} $v "" v
		}
		menu_addcommand $w $m $v "
			menu_setassoc $w fullname [list $orig] ; $prefix
		"
	}
}

proc menu_getassoc {w field} {
	global MenuData
	set data ""
	catch {set data $MenuData($w-$field)}
	return $data
}

proc menu_setassoc {w field data} {
	global MenuData
	set MenuData($w-$field) $data
}

proc menu_addcommand {menu num name {command ""}} {
	set m [menu_convertname $menu $num]
	$m add command -label $name -command $command
}

proc menu_changecommand {menu num name command} {
	set m [menu_convertname $menu $num]
	$m entryconfig $name -command $command
}

proc menu_addcheck {menu num name var} {
	set m [menu_convertname $menu $num]
	$m add check -label $name -var $var
}

proc menu_addsep {menu num} {
	set m [menu_convertname $menu $num]
	$m add separator
}

proc menu_addcascade {menu num name} {
	set m [menu_convertname $menu $num]
	$m add cascade -label $name
	set i [$m index $name]
	menu $m.m$i
	$m entryconfig $i -menu $m.m$i
}

proc optbtn_create {w name proc args} {
	frame $w
	label $w.label -borderwidth 0 -text $name
	entry $w.value -relief sunken -background white
	bind $w.value <Return> "$proc"
	menu_create $w.menu v

	pack $w.label -side left -fill both
	pack $w.value -side left -fill both -expand 1
	pack $w.menu -side right -fill both

	menu_setassoc $w.menu prefix \
		"optbtn_set $w \[menu_getassoc $w.menu fullname\] ; $proc"
	if {! [string match "" $args]} {
		eval menu_additems $w.menu $args
	}
}

proc optbtn_add {w args} {
	eval menu_additems $w.menu $args
}

proc optbtn_get {w} {
	return [$w.value get]
}

proc optbtn_set {w text} {
	$w.value delete 0 end
	$w.value insert 0 $text
}

proc placard_create {win image} {
	global PlacardMessage

        wm withdraw .
        toplevel $win -borderwidth 4 -relief raised
        wm overrideredirect $win 1
        window_center $win
        set img [image create photo -file $image]
        label $win.img -image $img
        pack $win.img -side top
	set PlacardMessage "Initializing..."
	label $win.l -textvar PlacardMessage
	pack $win.l -side bottom -fill x
        update
}

proc placard_label {msg} {
	global PlacardMessage
	set PlacardMessage $msg
	update
}

proc placard_destroy {w} {
	update
	destroy $w
}

proc radiobox_create {win {title ""}} {
	global rbInfo
	set rbInfo($win-current) ""
	set rbInfo($win-count) 0

	frame $win -class Radiobox
	if {$title != ""} {
		label $win.title -text $title
		pack $win.title -side top -anchor w
	}
	frame $win.border -borderwidth 2 -relief groove
	pack $win.border -expand yes -fill both
	return $win
}

proc radiobox_add {win choice {command ""}} {
	global rbInfo

	set name "$win.border.rb[incr rbInfo($win-count)]"
	radiobutton $name -text $choice -command $command \
		-variable rbInfo($win-current) -value $choice
	pack $name -side top -anchor w

	if {$rbInfo($win-count) == 1} {
		$name invoke
	}
}

proc radiobox_select {win choice} {
	global rbInfo
	set rbInfo($win-current) $choice
}

proc radiobox_get {win} {
	global rbInfo
	return $rbInfo($win-current)
}

global MessageData

proc message_create {w {msg ""}} {
	global MessageData
	set MessageData($w) $msg
	message $w -textvar MessageData($w) -width 5000
}

proc message_set {w {msg ""}} {
	global MessageData
	set MessageData($w) $msg
}

proc label_create {w label} {
	label $w -font -Adobe-times-medium-r-normal--*-180* -text $label
}

proc sep_create {w} {
	frame $w -height 2 -borderwidth 1 -relief sunken
}

proc checkbutton_create {w text} {
	global cbInfo
	checkbutton $w -text $text -variable cbInfo($w)
}

proc checkbutton_get {w} {
	global cbInfo
	return $cbInfo($w)
}
global text_StdTags
set text_StdTags(grey) "-background black -borderwidth 0 -bgstipple gray25"
set text_StdTags(reverse) "-background black -foreground white"
set text_StdTags(bold) "-font -*-Courier-Bold-R-Normal--*-120-*-*-*-*-*-*"
set text_StdTags(underline) "-underline 1"
set text_StdTags(yellow) "-background yellow"
set text_StdTags(blue) "-background blue"

proc text_create {w} {
	frame $w
	text $w.t -relief raised -bd 2 -yscrollcommand "$w.s set" \
		-setgrid true -width 80 -height 35 -cursor "" \
		-font -Adobe-Courier-Medium-R-Normal-*-120-*
	scrollbar $w.s -command "$w.t yview"
	pack $w.s -side right -fill y
	pack $w.t -expand yes -fill both

	bindtags $w.t "$w.t TextBrowse [winfo toplevel $w.t] all"
	bind $w.t <1> {
		%W tag remove sel 0.0 end
		%W mark set anchor [%W index @%x,%y]
	}
	bind $w.t <B1-Motion> {
		set tkPriv(x) %x
		set tkPriv(y) %y
		tkTextSelectTo %W %x %y
	}
	$w.t config -insertontime 0

	return $w
}

proc text_subwidget {w name} {
	switch $name {
		text	{return $w.t}
		scroll	{return $w.s}
	}
}

proc text_add {w text} {
	$w.t insert end $text
}

proc text_setfile {w {file ""}} {
	if {$file != ""} {
		if [file exists $file] {
			$w.t delete 0.0 end
			$w.t insert end [exec cat $file]
		} else {
			error "No such file: $file"
		}
	} else {
		$w.t delete 0.0 end
	}
}

proc text_set {w text} {
	$w.t delete 0.0 end
	$w.t insert 0.0 $text
}

proc text_get {w what} {
	switch $what {
		all	{return [$w.t get 0.0 end-1c]}
		selline	{
			set s [lindex [$w.t tag nextrange sel 0.0] 0]
			if {$s == ""} {return ""}
			return [$w.t get "$s linestart" "$s lineend"]
		}
		curline	{
			set x [expr [winfo pointerx $w] - [winfo rootx $w]]
			set y [expr [winfo pointery $w] - [winfo rooty $w]]
			set l [text_linenum $w $x $y]
			return [$w.t get $l.0 $l.end]
		}
		curword	{
			set x [expr [winfo pointerx $w] - [winfo rootx $w]]
			set y [expr [winfo pointery $w] - [winfo rooty $w]]
			return [$w.t get [$w.t index "@$x,$y wordstart"] \
				[$w.t index "@$x,$y wordend"]]
		}
		sel {
			set r [$w.t tag nextrange sel 0.0]
			if {$r != ""} {return [eval $w.t get $r]}
		}
	}
}

proc text_showline {w index} {
	focus $w.t
	$w.t see 0.0 	; # this will cause the real index to be centered
	$w.t see $index
}

proc text_numberlines {w} {

	set end [$w.t index end]
	for {set i 1} {$i < $end} {incr i} {
		set n [format %3d $i]
		$w.t insert $i.0 "$n "
	}
}

proc text_maketag {w name tag} {
	global text_StdTags
	catch {
		if {! [string match "" $text_StdTags($tag)]} {
			set tag $text_StdTags($tag)
		} 
	}
	eval $w.t tag config $name $tag
	$w.t tag raise $name
	$w.t tag raise sel
}

proc text_tagline {w i args}	{
	foreach tag $args {
		$w.t tag add $tag $i.0 "$i.end+1c"
	}
}

proc text_untagline {w i args} {
	foreach tag $args {
		if {$i == 0} {
			$w.t tag remove $tag 0.0 end
		} else {	
			$w.t tag remove $tag $i.0 "$i.end+1c"
		}
	}
}

proc text_tagcopy {w dst src} {
	if {[$w.t tag ranges $src] != ""} {
		eval [list $w.t tag add $dst] [$w.t tag ranges $src]
	}
}

proc text_hastag {w ind tag} {
	foreach i [$w.t tag names $ind] {
		if {$i == $tag} {return 1}
	}
	return 0
}

proc text_linenum {w {x ""} {y ""}} {
	if {$x == ""} {
		set index [$w.t index end-1c]
	} {
		set index [$w.t index @$x,$y]
	}
	return [lindex [split $index .] 0]
}

proc text_charnum {w {x ""} {y ""}} {
	if {$x == ""} {
		set index [$w.t index end-1c]
	} {
		set index [$w.t index @$x,$y]
	}
	return [lindex [split $index .] 1]
}
global mkWindowNum
set mkWindowNum 1

proc window_create {{v ""} {opt ""}} {
	global mkWindowNum
	while {[winfo exists .win$mkWindowNum]} {
		incr mkWindowNum
	}

	set w .win$mkWindowNum
	if {$opt != ""} {
		toplevel $w -class $opt
	} else {
		toplevel $w 
	}
	set mkWindowName $w
	if {! [string match $v ""]} {uplevel "set $v $w"}
	return $w
}

proc window_title {win title {icontitle ""}} {
	wm title [winfo toplevel $win] $title
	if {! [string match $icontitle ""]} {
		wm iconname [winfo toplevel $win] $icontitle
	}
}

proc window_center {win} {
        after idle "
                update idletasks
                set xmax \[winfo screenwidth $win\]
                set ymax \[winfo screenheight $win\]
                set x0 \[expr (\$xmax-\[winfo reqwidth $win\])/2\]
                set y0 \[expr (\$ymax-\[winfo reqheight $win\])/2\]
                wm geometry $win +\$x0+\$y0"
}

proc window_setassoc {w key data} {
	global mkWindowData_$w
	set mkWindowData_$w($key) $data
}

proc window_getassoc {w key} {
	global mkWindowData_$w
	return [set mkWindowData_$w($key)]
}

proc window_grab {w} {
	wm protocol $w WM_DELETE_WINDOW { }
	grab $w
	focus $w
}

proc window_ungrab {w} {
	grab release $w
}

proc window_transient {w master} {
	wm transient $w [winfo toplevel $master]
}

proc window_pos {w} {
	return [list	[expr [winfo pointerx $w] - [winfo rootx $w]] \
			[expr [winfo pointery $w] - [winfo rooty $w]]]
}

proc window_list {class} {
	set result ""
	foreach w [winfo children .] {
		if {[winfo class $w] == $class} {lappend result $w}
	}
	return $result
}

proc options_translate {opt} {
	set i 0
	set flag ""
	while {1} {
		set c [string index $opt $i]
		incr i
		switch $c {
			""	{ return "$flag" }
			t	{append flag " -side top"}
			b	{append flag " -side bottom"}
			l	{append flag " -side left"}
			r	{append flag " -side right"}
			e	{append flag " -expand yes"}
			x	{append flag " -fill x"}
			y	{append flag " -fill y"}
			z	{append flag " -fill both"}
			4	{append flag " -padx 4 -pady 4"}
			P	{append flag " -padx 20 -pady 3"}
		}
	}
}

proc mpack {args} {
	foreach l $args {
		set frame [lindex [split $l :] 0]
		set opt   [lindex [split $l :] 1]
		if [string match $opt ""] {set opt "tx"}
		set flag [options_translate $opt]
		eval pack $frame $flag
	}
}
